[TOC]

# XV6文件系统模块设计文档 

## 一、getcwd系统调用

### 1.1 getcwd功能说明

在该系统中inode的inum是文件唯一的标识符号，所以通过对比inum得到文件所在的路径。

这里使用dfs来查询，找到文件的路径信息。
得到的文件信息通过传入的char*参数返回，copyout写回进程的用户页表。

这里的细节是，不能创造char数组再赋予返回指针，因为只是在函数里面，函数执行完成会回收空间。

又由于inum是唯一的，所以当内存内的inode得到inum的时候并没有导入磁盘内部的inode的数据，这里`ilock(dp)`,不仅仅是上锁，更是导入磁盘内部的数据。

### 1.2 getcwd关键代码

```c
static uint64 dfs(struct inode *dp,ushort inum,char* kbuf)//搜索路径
{    begin_op();
  ilock(dp);
  if(dp->type != T_DIR)//不是dir
   {iunlockput(dp);
    end_op();
    return -1;}
  uint off;
  struct dirent de;
  for(off = 0; off < dp->size; off += sizeof(de)){
    if(readi(dp, 0, (uint64)&de, off, sizeof(de)) != sizeof(de))
      panic("dirlookup read");
    //printf("%s %d\n",de.name,de.inum);
    if (strncmp(de.name,".",strlen(de.name))==0||strncmp(de.name,"..",strlen(de.name))==0)//遍历跳过这两文件
    {continue;}
    if(de.inum == inum)
    { strncpy(kbuf,"/",1);
      strncpy(kbuf+strlen(kbuf),de.name,strlen(de.name)+1);
      iunlockput(dp);
      end_op();
      return 0;}
    else{
      strncpy(kbuf,"/",1);
      strncpy(kbuf+strlen(kbuf),de.name,strlen(de.name)+1);
      int re = dfs(dirlookup(dp,de.name,0),inum,(kbuf+strlen(de.name)+1));
      if(re==0)
      {iunlockput(dp);
        end_op();
        return 0;}
      else{memset(kbuf+1,0,strlen(kbuf+1));
      continue;}
    }
    }
    iunlockput(dp);
    end_op();
  return -1;
}

uint64 sys_getcwd()
{ 
      uint64 buf; // 用户空间缓冲区的地址
  int n;     // 缓冲区的大小
  argaddr(0, &buf);
  argint(1, &n);
  char kbuf[n];
  struct proc* p = myproc();
  struct inode *dp,*ip;
  ip = idup(myproc()->cwd);
  dp = idup(initproc->cwd);
  //printf("inum:%d %d\n",dp->inum,ip->inum);
  strncpy(kbuf,"/",2);
  if(ip->inum==1){//这是在根目录里面
    if(copyout(p->pagetable, buf, kbuf, strlen(kbuf) + 1) < 0)
       return -1; // 复制失败
    return 0;
  }

  //开始找寻目录路径
  uint off;
  struct dirent de;
  begin_op();
  ilock(dp);
  for(off = 0; off < dp->size; off += sizeof(de)){
    if(readi(dp, 0, (uint64)&de, off, sizeof(de)) != sizeof(de))
      panic("dirlookup read");
    //printf("%s %d %d\n",de.name,de.inum,dp->size);
    if (strncmp(de.name,".",strlen(de.name))==0||strncmp(de.name,"..",strlen(de.name))==0)//遍历跳过这两文件
    {continue;}
    if(de.inum == ip->inum)
    { strncpy(kbuf+strlen(kbuf),de.name,strlen(de.name)+1);
      iunlockput(dp);
      end_op();
      if(copyout(p->pagetable, buf, kbuf, strlen(kbuf) + 1) < 0)
       return -1; // 复制失败
      return 0;}
    else{
      strncpy(kbuf+strlen(kbuf),de.name,strlen(de.name));//加1是因为上面加入了‘/’
      int re = dfs(dirlookup(dp,de.name,0),ip->inum,(kbuf+strlen(de.name)+1));
      if(re==0)
      {iunlockput(dp);
        end_op();
        if(copyout(p->pagetable, buf, kbuf, strlen(kbuf) + 1) < 0)
       return -1; // 复制失败
        return 0;}
      else{memset(kbuf+1,0,strlen(kbuf+1));
      continue;}
    }
    }
    iunlockput(dp);
    end_op();
    if(copyout(p->pagetable, buf, kbuf, strlen(kbuf) + 1) < 0)
       return -1; // 复制失败
  return -1 ;
}
```

### 1.3 getcwd测试结果

下面是运行的代码和结果：

```c
#include "kernel/types.h"
#include "kernel/stat.h"
#include "user/user.h"
int main ()
{ 
char buf[20];
printf("%ld\n",(uint64)buf);
getcwd(buf,20);
printf("test:%s\n",buf);
return 0;
}
```
![Local Image](./figs/getcwd.png)


## 二、文件相关命令

### 2.1 rename命令
一开始，仅仅认为rename把目录文件读出，然后修改名字后在写回磁盘，就实现的rename功能，形状是有了，但后来写mv的时候改正了该问题。
mv本来用link和unlink实现，但发现，link和unlink不能操作目录，因为可能形成目录的循环，你中有我，我中有你的困境等问题，所以使用rename，
这里的rename抄写了unlink和link，过滤了他们对目录文件的争对，变成一个系统调用，rename的改名字功能仍然存在。

```c
uint64 sys_rename()
{ char name[DIRSIZ], new[MAXPATH], old[MAXPATH];
  struct inode *dp, *ip;
  if(argstr(0, old, MAXPATH) < 0 || argstr(1, new, MAXPATH) < 0)
    return -1;

  begin_op();
  if((ip = namei(old)) == 0){
    end_op();
    return -1;
  }
  ilock(ip);
  ip->nlink++;
  iupdate(ip);
  iunlock(ip);

  if((dp = nameiparent(new, name)) == 0)
    {
      goto bad;}
  ilock(dp);
  if(dp->dev != ip->dev || dirlink(dp, name, ip->inum) < 0){
    iunlockput(dp);
    goto bad;
  }
  iunlockput(dp);
  iput(ip);
  //上面是模仿link开始模仿unlink部分
  struct dirent de;
  uint off;
  if((dp = nameiparent(old, name)) == 0){
    end_op();
    return -1;
  }

  ilock(dp);

  // Cannot unlink "." or "..".
  if(namecmp(name, ".") == 0 || namecmp(name, "..") == 0)
    goto ubad;

  if((ip = dirlookup(dp, name, &off)) == 0)
    goto ubad;
  ilock(ip);

  if(ip->nlink < 1)
    panic("unlink: nlink < 1");

  memset(&de, 0, sizeof(de));
  if(writei(dp, 0, (uint64)&de, off, sizeof(de)) != sizeof(de))
    panic("unlink: writei");
  if(ip->type == T_DIR){
    dp->nlink--;
    iupdate(dp);
  }
  iunlockput(dp);

  ip->nlink--;
  iupdate(ip);
  iunlockput(ip);

  end_op();
  return 0;

ubad:
printf("ubad\n");
  iunlockput(dp);
  end_op();
  return -1;
bad:
printf("bad\n");
  ilock(ip);
  ip->nlink--;
  iupdate(ip);
  iunlockput(ip);
  end_op();
  return -1;
}
```

### 2.2 mv命令

这是mv的实现

```c
#include "kernel/types.h"
#include "kernel/stat.h"
#include "user/user.h"
int main(int argc,char *argv[]) {
    
    printf("%s\n%s\n",argv[1],argv[1]);
    if(argv[1][0]!='/'||argv[2][0]!='/')
    {
    char str1[2+strlen(argv[1])+1];
    char str2[2+strlen(argv[2])+1];
    if(argv[1][0]=='/')
    {
    strcpy(str1,argv[1]);}
    else{
    strcpy(str1,".");
    strcpy(str1+strlen(str1),"/");
    strcpy(str1+strlen(str1),argv[1]);
    }
     if(argv[2][0]=='/')
    {
    strcpy(str2,argv[2]);}
    else{
    strcpy(str2,".");
    strcpy(str2+strlen(str2),"/");
    strcpy(str2+strlen(str2),argv[2]);
    }
    printf("%s %s\n",str1,str2);
    rename(str1,str2);
    // link(str1,str2);
    // unlink(str1);
    
    }
    else{
        printf("%s %s\n",argv[1],argv[2]);
        rename(argv[1],argv[2]);
        // link(argv[1],argv[2]);
        // unlink(argv[1]);
    }
    
 
    return 0;
}
```

下面是运行的代码和结果：

```c
#include "kernel/types.h"
#include "kernel/stat.h"
#include "user/user.h"
int main ()
{ 
int a;
char buf[100] = "nihaoaaaa";
printf("%ld\n",(uint64)buf);
a = open("aaatest",0x200| 0x001);
write(a,buf,strlen(buf));
return 0;
}
```


![Local Image](./figs/mv.png)


### 2.3 rmdir命令

在文件系统中，删除目录是一个基础但关键的操作。rmdir系统调用用于删除空目录，这是文件系统管理的重要组成部分。该操作需要保证目录为空且维护文件系统的一致性。

rmdir命令通过以下步骤实现目录删除：
```c
uint64
sys_rmdir(void)
{
  char path[MAXPATH], name[DIRSIZ];
  struct inode *ip, *dp;
  uint off;
  struct dirent de;

  begin_op();  // 开始文件系统操作
  // 获取路径并检查目录是否存在
  if(argstr(0, path, MAXPATH) < 0 || (ip = namei(path)) == 0){
    end_op();
    return -1;
  }
  
  // 检查目录有效性：必须是目录类型、链接数为1且为空
  ilock(ip);
  if(ip->type != T_DIR || ip->nlink > 1 || !isdirempty(ip)){
    iunlockput(ip);
    end_op();
    return -1;
  }
  
  // 清除目录项并更新父目录和当前目录的链接计数
  dp = nameiparent(path, name);
  // ...更新文件系统结构...
  end_op();
  return 0;
}
```

rmdir采用了严格的安全检查机制，确保只能删除空目录，避免误删非空目录。使用begin_op/end_op事务机制保证文件系统操作的原子性，即使在系统崩溃时也能保持一致性。通过维护目录项和链接计数，正确处理文件系统的层次结构关系。

使用示例：
```bash
$ mkdir test          # 创建测试目录
$ rmdir test          # 删除空目录
$ rmdir test          # 如果目录不存在，返回错误
$ mkdir test          
$ touch test/file     # 创建文件
$ rmdir test          # 删除非空目录将失败
```

这个设计既保证了操作的安全性，又维护了文件系统的完整性，是一个稳健的系统调用实现。


## 三、一些小细节

### 3.1 文件的执行

使用c语言&&的性质，当在某一文件目录下，找不到执行文件就会去根目录下查找，让ls和mv等命令可以到处执行，
但不足是，像Linux当前目录文件的执行是./文件名来实现，这里并没有修改,当然也是可以的，因为 `.`刚好指向当前目录的所在。
```c
  char fromroot[strlen(path+2)];
  strncpy(fromroot,"/",2);
  strncpy(fromroot+1,path,strlen(path)+1);
  if((ip = namei(path)) == 0&& (ip =namei(fromroot))==0){
    end_op();
    return -1;
  }
  ilock(ip);
```


### 3.2 路径的显示

要显示当前路径所在的位置，是控制台的重要功能。利用前面实现的`getcwd()`得到文件的路径情况。

```c
int
getcmd(char *buf, int nbuf, char *cwd)
{ 
  write(2,cwd, strlen(cwd));
  write(2,"$ ",2);
  memset(buf, 0, nbuf);
  gets(buf, nbuf);
  if(buf[0] == 0) // EOF
    return -1;
  return 0;
}

int
main(void)
{
  static char buf[100];
  static char cwd[100];//显示当前路径
  int fd;
  // Ensure that three file descriptors are open.
  while((fd = open("console", O_RDWR)) >= 0){
    if(fd >= 3){
      close(fd);
      break;
    }
  }
  strcpy(cwd,"/");
  // Read and run input commands.
  while(getcmd(buf, sizeof(buf),cwd) >= 0){
    if(buf[0] == 'c' && buf[1] == 'd' && buf[2] == ' '){
      // Chdir must be called by the parent, not the child.
      buf[strlen(buf)-1] = 0;  // chop \n
      if(chdir(buf+3) < 0)
        {fprintf(2, "cannot cd %s\n", buf+3);continue;}//代表cd失败
      getcwd(cwd,100);
      continue;
    }
    if(fork1() == 0)
      runcmd(parsecmd(buf));
    wait(0);
  }
  exit(0);
}
```
细节也在上面的用例上体现了，这里就不在举例了。

